/* -------------------------------- objects.c ------------------------------- */

/* This is part of the flight simulator 'fly8'.
 * Author: Eyal Lebedinsky (eyal@ise.canberra.edu.au).
*/

/* Some generic object behaviour routines and hit detection.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "fly.h"

/* hit routines.
*/
static void FAR
object_rand (OBJECT *obj, int speed, int extent, int noise)
{
	int	e, i;

	e =  SH(obj)->extent;
	if (!e)
		e = 1;
	if (extent/64 <= e) {
		if (0 == (i = SH(obj)->extent))
			i = 1;
		i = muldiv (speed/64, extent, i);
	} else
		i = speed;
	if (i) {
		obj->V[X] += rand () % i;
		obj->V[Y] += rand () % i;
		obj->V[Z] += rand () % i;
	}
	i = 0x7fff/e;
	if (i) {
		obj->speed = ihypot3d (obj->V);
		obj->da[X] = rand ()%i-i/2;
		obj->da[Y] = rand ()%i-i/2;
		obj->da[Z] = rand ()%i-i/2;
	}

	if (noise && st.quiet)
		Snd->Beep (1414, 50);
}

static int FAR
object_break (int n, int speed, int extent, int noise)
{
	int	i;
	OBJECT	*p;

	for (i = 0; i < n; ++i) {
		if (!(p = create_object (O_BROKEN, 1)))
			break;
		p->flags |= F_HIT;
		object_rand (p, speed, extent, noise);
	}
	return (i);
}

extern void FAR
object_hit (OBJECT *obj, int seed, int speed, int extent, int damaging)
{
	srand (seed);

	obj->damage -= damaging;

	if (obj == CC) {
		MsgWPrintf (50, "damage %d", obj->damage);
		if (st.quiet)
			Snd->List (TnDamage, 0);
	}

	st.bodies[obj->name]->hit (obj, speed, extent, damaging);

	if (obj->flags & F_EXPORTED)	/* exported victim */
		remote_imhit (obj, seed, speed, extent, damaging);
}

static int FAR
hit (VECT p, VECT d, int r)	/* initial distance p, closure speed r */
{
	long	x, y, z, s, t;

	x = p[X];
	y = p[Y];
	z = p[Z];

	s = x*d[X]+y*d[Y]+z*d[Z];
	if (s < 0)		/* objects getting away from each other */
		return (0);
	if (s > 0) {
		t = d[X]*(long)d[X]+d[Y]*(long)d[Y]+d[Z]*(long)d[Z];
		if (s >= t) {
#if 0
			x -= d[X];	/* not really needed now */
			y -= d[Y];
			z -= d[Z];
#endif
			return (0);	/* objects will be closer */
		} else {
			x -= d[X]*s/t;
			y -= d[Y]*s/t;
			z -= d[Z]*s/t;
		}
	}

#if 0
	return ((x*x+y*y+z*z) < r*(long)r);
#else
	return (est_hyp ((int)x, (int)y, (int)z) < (Uint)r);
#endif
}

static void FAR
check_hit (OBJECT *p, int interval)
{
	register int	hit_dist;
	register OBJECT	*w;
	int		t2, extent;
	long		t1, t3;
	VECT		r, d, V;
	LVECT		R;
	OBJECT		*owner;

	if (!(p->shflags & SH_HIT) &&
	    (st.flags & SF_BLANKER || !(p->flags & F_HIT)))
		return;

	if (p->flags & F_IMPORTED)		/* ignore foreign bullets */
		return;

	V[X] =  p->V[X]/VONE;
	V[Y] =  p->V[Y]/VONE;
	V[Z] =  p->V[Z]/VONE;

	R[X] =  p->R[X]/VONE;
	R[Y] =  p->R[Y]/VONE;
	R[Z] =  p->R[Z]/VONE;

	extent = SH(p)->extent;
	owner = p->owner;

	for (w = COT; w; w = w->prev) {
		if (!(w->shflags & SH_BEHIT))
			break;			/* all BEHIT's are at end! */
		if (w == p || (w == owner && w->id == p->ownerid))
			continue;
		if (w->flags & F_DEL)
			continue;

		hit_dist = fmul (FONE/4*3, SH(w)->extent + extent);

		t1 = w->R[X]/VONE - R[X];
		t2 = TADJ(w->V[X]/VONE - V[X]);
		t3 = t1 - t2;
		if (t1 >  hit_dist && t3 >  hit_dist)
			continue;
		if (t1 < -hit_dist && t3 < -hit_dist)
			continue;
		r[X] = (int)t1;
		d[X] = t2;

		t1 = w->R[Y]/VONE - R[Y];
		t2 = TADJ(w->V[Y]/VONE - V[Y]);
		t3 = t1 - t2;
		if (t1 >  hit_dist && t3 >  hit_dist)
			continue; 
		if (t1 < -hit_dist && t3 < -hit_dist)
			continue;
		r[Y] = (int)t1;
		d[Y] = t2;

		t1 = w->R[Z]/VONE - R[Z];
		t2 = TADJ(w->V[Z]/VONE - V[Z]);
		t3 = t1 - t2;
		if (t1 >  hit_dist && t3 >  hit_dist)
			continue;
		if (t1 < -hit_dist && t3 < -hit_dist)
			continue;
		r[Z] = (int)t1;
		d[Z] = t2;

		if (hit (r, d, hit_dist)) {
			if (!(w->flags & F_HIT) &&
				 owner && owner->id == p->ownerid)
					++p->owner->score;
			p->flags |= F_HIT|F_MOD;
			p->color = st.white;
			if (w->flags & F_IMPORTED)	/* imported victim */
				remote_urhit (w, p->speed, extent,
					p->damaging);
			else				/* local victim */
				object_hit (w, rand(), p->speed, extent,
					p->damaging);
		}
	}
}

/* Some generic routines for bodies.
*/

extern int FAR
body_init (BODY *b)
{
	return (0);
}

extern void FAR
body_term (BODY *b)
{
	return;
}

extern void FAR
body_delete (OBJECT *p)
{
	return;
}

extern void FAR
body_dynamics (OBJECT *p, int interval)
{
	if (((st.flags1 & SF_USEG) && (p->shflags & SH_G)) ||
	    (p->flags & F_HIT)) {
		if (iabs (p->speed) < (Uint)(VMAX-10*GACC)) {
			p->V[Z] -= TADJ (GACC);
			p->speed = ihypot3d (p->V);
		}
	}

	if (p->flags & F_HIT) {
		if (FOREVER == p->time)
			p->color = rand()%2 ? st.red : st.white;
		else
			p->color = (p->time/(TIMEPSEC/10))&1
						? st.red : st.white;
		if (p->name == O_PLANE) {
			st.owner = p;
			object_break (rand()%2, p->speed, 1, 0);
		}
		if (!(p->flags & F_IMPORTED) &&
		    p->time == FOREVER && p->R[Z] <= 0) {
			p->flags |= F_DEL|F_MOD;
			if (p->name == O_PLANE) {
				st.owner = p;
				object_break (rand()%16+5, p->speed,
					SH(p)->extent, 1);
			}
		}
	}

	if (p->da[X] || p->da[Y] || p->da[Z]) {
		p->a[X] += TADJ(p->da[X])*VONE;
		if (iabs(p->a[X]) > D90) {
			p->a[X] = D180 - p->a[X];
			p->a[Y] = D180 + p->a[Y];
			p->da[X] = -p->da[X];
		}
		p->a[Y] += TADJ(p->da[Y])*VONE;
		p->a[Z] += TADJ(p->da[Z])*VONE;
		Mobj (p);
	}
}

extern void FAR
body_hit (OBJECT *obj, int speed, int extent, int damaging)
{
	if (obj->damage <= 0 && !(obj->flags & F_HIT)) {
		obj->color = st.white;
		obj->flags |= F_HIT|F_MOD;
		if (obj->name == O_PLANE) {
			obj->time = FOREVER;
			if (obj == CC && !(EE(obj)->hud1&HUD_VALARM))
				MsgEPrintf (50, "YOU'RE HIT, EJECT!");
			if (obj->damage == 0 && !(obj->flags & F_IMPORTED) &&
			    !(obj->flags&F_CC))
				eject (obj);
		} else
			obj->time = 5*TIMEPSEC;
	}

	if (obj->name == O_TARGET || obj->name == O_GTARGET)
		obj->flags |= F_DEL;	/* no notification */
/* explode.
*/
	st.owner = obj;
	if (obj->damage >= -20) {	/* limit vicious loops */
		object_break (rand()%5 + 2, speed, extent, 0);
	}

	object_rand (obj, speed, extent, 1);
}

/* objects handling.
*/

#define NM	(1852*VONE)

static void FAR
object_update (OBJECT *p, int interval)
{
	int	tx, ty;

	p->R[X] += (tx = TADJ (p->V[X]));
	p->R[Y] += (ty = TADJ (p->V[Y]));
	p->R[Z] += TADJ (p->V[Z]);

	if (p->flags & F_KEEPNAV) {
		if (p->misc[3] >= 90*60 || p->misc[3] < -90*60)
			tx = -tx;
		p->misc[2] += tx;
		tx = fmul (NM, COS(muldiv(p->misc[3],D90,90*60)));
		while (p->misc[2] >= tx) {
			p->misc[2] -= tx;
			if (++p->misc[1] >= 180*60)
				p->misc[1] -= 360*60;
		}
		while (p->misc[2] < 0) {
			p->misc[2] += tx;
			if (--p->misc[1] < -180*60)
				p->misc[1] += 360*60;
		}

		p->misc[4] += ty;
		while (p->misc[4] >= NM) {
			p->misc[4] -= NM;
			if (++p->misc[3] >= 180*60)
				p->misc[3] -= 180*60;
		}
		while (p->misc[4] < 0) {
			p->misc[4] += NM;
			if (--p->misc[3] < -180*60)
				p->misc[3] += 180*60;
		}
	}

	if (p->time != FOREVER) {
		p->time -= muldiv (interval, TIMEPSEC, 1000);
		if (p->time <= 0)
			p->flags |= F_DEL;	/* no notification */
	}

	p->flags &= ~F_DONE;
}

static void FAR
object_dynamics (OBJECT *p, int interval)
{
	BODY	*b;
	int	rinterval;

	if (p->flags & F_DONE)
		return;

	st.owner = p;
	b = st.bodies[p->name];

	if (p->flags & F_IMPORTED) {
		remote_receive (p);
		randomize ();		/* the net syncs the machines */
		if (ET_IMPORTED == p->e_type && EIM(p)->timeout < st.present) {
			p->flags |= F_DEL;		/* object timeout */
			++st.stats[51];
		} else {
			rinterval = (int)(st.present - p->rtime);
			if (rinterval > 0) {
				if (rinterval > 1000)	/* ignore long pauses */
					rinterval = st.interval;
				if (b->shape->flags & SH_LOCALSIM)
					b->dynamics (p, rinterval);
				else
					body_dynamics (p, rinterval);
			}
			p->rtime = st.present;
		}
	} else {
		b->dynamics (p, interval);
		check_hit (p, interval);
	}
	p->flags |= F_DONE;
}

extern int FAR
objects_dynamics (int interval)
{
	OBJECT	*p, *q;
	long	t;
	int	refreshed;

	remote_receive (0);
	randomize ();		/* the net syncs the machines */

	for (p = CO; p; p = p->next)
		object_update (p, interval);

	t = Tm->Milli ();
	for (p = CO; p; p = p->next) {
		if ((q = p->owner) && p->ownerid == q->id)
			object_dynamics (q, interval);
		object_dynamics (p, interval);
	}
	st.stats[28] += Tm->Milli () - t;

	remote_refresh ();
	players_flush ();	/* del bad messages */
	players_purge ();	/* del bad players */

	if (CC->flags & F_DEL) {
		p = CC;
		st.options = st.ptype;
		if (!(CC = create_object (O_PLANE, 1))) {
			MsgEPrintf (-50, "no new plane");
			return (1);
		}
		MsgPrintf (50, "new plane");
		if ((CC->pointer = p->pointer))
			(*CC->pointer->control->Center)(CC->pointer);
		EE(CC)->flags |= EE(p)->flags;
		EE(CC)->hud  = EE(p)->hud;
		EE(CC)->hud1 = EE(p)->hud1;
		EE(CC)->hud2 = EE(p)->hud2;
		EE(CC)->radar = EE(p)->radar;
		EE(CC)->hudarea = EE(p)->hudarea;
		EE(CC)->tapelen = EE(p)->tapelen;
		EE(CC)->weapon = EE(p)->weapon;
		EE(CC)->misc[17] = EE(p)->misc[17];
		CC->flags |= p->flags & ~(F_HIT|F_DEL);
		CC->gpflags = p->gpflags;
		if (!(CC->gpflags & GPF_PILOT))
			CC->flags |= F_STEALTH;		/* wait for pilot */
		CC->score  = p->score;
		p->pointer = 0;
		CC->viewport = p->viewport;
		p->viewport = 0;
		if (st.flags & SF_BLANKER)
			EE(CC)->weapon = WE_M61;
		for (q = CO; q; q = q->next) {
			if (O_BROKEN == q->name && q->owner == p) {
				q->owner = CC;
				q->ownerid = CC->id;
			}
		}
	}

	if (CV->flags & F_DEL) {
		if ((p = CV->owner) && p->id == CV->ownerid &&
		    !(p->flags & F_DEL)) {
			CV = p;
		} else
			CV = CC;
		if (CV->viewport)
			get_viewport (CV);
	}

/* export, refresh and delete as needed.
*/
	refreshed = 0;
	st.ntargets = 0;
	for (p = CO; p;) {
		if (p->flags & F_EXPORTED) {
			if (!refreshed && !(p->flags&(F_DEL|F_NEW|F_ALIVE))) {
				p->flags |= F_MOD|F_ALIVE;
				refreshed = 1;
			}
			send_obj (p, 0, 3);
		}

		if (p->flags & F_DEL) {
			p = delete_object (p);
			continue;
		}

		if ((p->name == O_TARGET || p->name == O_GTARGET)
		    && !(p->flags & F_HIT))
			++st.ntargets;
		p->flags &= ~(F_NEW|F_MOD);
		p = p->next;
	}

	if (!refreshed && st.present > st.RefreshTime) {	/* restart */
		for (p = CO; p; p = p->next)
			if (p->flags & F_EXPORTED)
				p->flags &= ~F_ALIVE;
		st.RefreshTime = st.present + TO_REFRESH;
	}

	send_obj (0, 0, 0);		/* flush remote */
	return (0);
}
